# =============================================================================
# Pelote Tabular to Network Conversion Functions
# =============================================================================
#
# Functions able to convert tabular data to networkx graphs.
#
import networkx as nx
from typing import Sequence, Union, Callable, Dict, Tuple, Any, Optional

from pelote.utils import IncrementalIdRegister
from pelote.shim import is_dataframe
from pelote.types import AnyGraph, Tabular, GenericKey, Indexable

RowDataSpec = Union[Sequence[GenericKey], Callable[[Indexable], Dict[Any, Any]]]


def collect_row_data(spec: RowDataSpec, row: Indexable) -> Dict[Any, Any]:
    if callable(spec):
        attr = spec(row)

        if not isinstance(attr, dict):
            raise TypeError(
                "row data collection should return a dict but returned %s instead"
                % type(attr).__name__
            )

        return attr

    return {k: row[k] for k in spec}


def table_to_bipartite_graph(
    table: Tabular,
    first_part_col: GenericKey,
    second_part_col: GenericKey,
    *,
    node_part_attr: str = "part",
    edge_weight_attr: str = "weight",
    first_part_data: Optional[RowDataSpec] = None,
    second_part_data: Optional[RowDataSpec] = None,
    disjoint_keys: bool = False
) -> AnyGraph:
    """
    Function creating a bipartite graph from the given tabular data.

    Args:
        table (Iterable[Indexable] or pd.DataFrame): input tabular data. It can
            be a large variety of things as long as it is 1. iterable and 2.
            yields indexable values such as dicts or lists. This can for instance
            be a list of dicts, a csv.DictReader stream etc. It also supports
            pandas DataFrame if the library is installed.
        first_part_col (str or int): the name of the column containing the
            value representing a node in the resulting graph's first part.
            It could be the index if your rows are lists or a key if your rows
            are dicts instead.
        second_par_col (str or int): the name of the column containing the
            value representing a node in the resulting graph's second part.
            It could be the index if your rows are lists or a key if your rows
            are dicts instead.
        node_part_attr (str, optional): name of the node attribute containing
            the part it belongs to. Defaults to "part".
        edge_weight_attr (str, optional): name of the edge attribute containing
            its weight, i.e. the number of times it was found in the table.
            Defaults to "weight".
        first_part_data (Sequence or Callable, optional): sequence (i.e. list, tuple etc.)
            of column from rows to keep as node attributes for the graph's first part.
            Can also be a function returning a dict of those attributes.
            Note that the first row containing a given node will take precedence over
            subsequent ones regarding data to include.
            Defaults to None.
        second_part_data (Sequence or Callable, optional): sequence (i.e. list, tuple etc.)
            of column from rows to keep as node attributes for the graph's second part.
            Can also be a function returning a dict of those attributes.
            Note that the first row containing a given node will take precedence over
            subsequent ones regarding data to include.
            Defaults to None.
        disjoint_keys (bool, optional): set this to True as an optimization
            mechanism if you know your part keys are disjoint, i.e. if no
            value for `first_part_col` can also be found in `second_part_col`.
            If you enable this option wrongly, the result can be incorrect.
            Defaults to False.
    """

    if first_part_col == second_part_col:
        raise TypeError("first_part_col and second_part_col must be different")

    if is_dataframe(table):
        table = (row for _, row in table.iterrows())

    graph = nx.Graph()
    node_id = IncrementalIdRegister[Tuple[GenericKey, Any]]()

    for i, row in enumerate(table):
        try:
            label1 = row[first_part_col]
            label2 = row[second_part_col]
        except (IndexError, KeyError):
            raise TypeError(
                'row %i lacks the "%s" or the "%s" value'
                % (i, first_part_col, second_part_col)
            )

        if disjoint_keys:
            n1 = label1
            n2 = label2
        else:
            # TODO: possibility to save lookups for sorted data
            n1 = node_id[first_part_col, label1]
            n2 = node_id[second_part_col, label2]

        if n1 not in graph:
            node_attr = {node_part_attr: first_part_col, "label": str(label1)}

            if first_part_data:
                node_attr.update(collect_row_data(first_part_data, row))

            graph.add_node(n1, **node_attr)

        if n2 not in graph:
            node_attr = {node_part_attr: second_part_col, "label": str(label2)}

            if second_part_data:
                node_attr.update(collect_row_data(second_part_data, row))

            graph.add_node(n2, **node_attr)

        if graph.has_edge(n1, n2):
            graph[n1][n2][edge_weight_attr] += 1
        else:
            edge_attr = {edge_weight_attr: 1}
            graph.add_edge(n1, n2, **edge_attr)

    return graph



# =============================================================================
# Pelote Network Projection Functions
# =============================================================================
#
# Functions used to compute various networkx projections such as bipartite
# to monopartite, for instance.
#
import math
import networkx as nx
from typing import Hashable, Optional, Counter, Any, Dict
from typing_extensions import Literal

from pelote.types import AnyGraph
from pelote.graph import check_graph
from pelote.utils import dict_without

MONOPARTITE_PROJECTION_METRICS = {
    "jaccard",
    "overlap",
    "cosine",
    "dice",
    "binary_cosine",
}
MonopartiteProjectionMetric = Literal[
    "jaccard", "overlap", "cosine", "dice", "binary_cosine"
]


def compute_metric(
    metric: Optional[MonopartiteProjectionMetric],
    norm1: float,
    norm2: float,
    i: float,
):
    if i == 0:
        return 0

    if metric == "cosine":
        return i / (norm1 * norm2)

    if metric == "jaccard":
        return i / (norm1 + norm2 - i)

    if metric == "dice":
        return (2 * i) / (norm1 + norm2)

    if metric == "overlap":
        return i / min(norm1, norm2)

    if metric == "binary_cosine":
        return i / math.sqrt(norm1 * norm2)

    return i


# TODO: bipartition check (is_bipartite_by, graph.py)
# TODO: partition is not empty
def monopartite_projection(
    bipartite_graph: AnyGraph,
    part_to_keep: Hashable,
    *,
    node_part_attr: str = "part",
    edge_weight_attr: str = "weight",
    metric: Optional[MonopartiteProjectionMetric] = None,
) -> AnyGraph:

    check_graph(bipartite_graph)

    if metric is not None and metric not in MONOPARTITE_PROJECTION_METRICS:
        raise TypeError(
            'unknown metric "%s", expecting one of %s'
            % (metric, ", ".join('"%s"' % m for m in MONOPARTITE_PROJECTION_METRICS))
        )

    monopartite_graph = nx.Graph()

    # Computing norms
    norms: Dict[Any, float] = {}

    for n1, a1 in bipartite_graph.nodes(data=True):
        p1 = a1[node_part_attr]

        if p1 != part_to_keep:
            continue

        norm: float = 0

        for token, ta in bipartite_graph[n1].items():
            weight = 1

            if metric == "cosine":
                weight = ta[edge_weight_attr]
                weight *= weight

            norm += weight

        if norm > 0:
            if metric == "cosine":
                norm = math.sqrt(norm)

            norms[n1] = norm

        monopartite_graph.add_node(n1, **dict_without(a1, node_part_attr))

    for n1, norm1 in norms.items():
        intersection: Counter[Any] = Counter()

        # Computing intersections
        for token, ta in bipartite_graph[n1].items():
            w1 = 1

            if metric == "cosine":
                w1 = ta[edge_weight_attr]

            for n2, a2 in bipartite_graph[token].items():

                # Don't compare to self
                if n2 == n1:
                    continue

                # NOTE: since we are dealing with undirected graphs we can
                # avoid doing the same computations twice.
                if n1 > n2:
                    continue

                w2 = 1

                if metric == "cosine":
                    w2 = a2[edge_weight_attr]

                intersection[n2] += w1 * w2

        # Finalizing metrics
        for n2, i in intersection.items():
            norm2 = norms[n2]
            weight = compute_metric(metric, norm1, norm2, i)

            # TODO: threshold can go there
            monopartite_graph.add_edge(n1, n2, weight=weight)

    return monopartite_graph



# =============================================================================
# Pelote Learning Routines
# =============================================================================
#
# Collection of learning function related to graph sparsification mostly.
#
import math
from typing import Optional

from pelote.types import AnyGraph
from pelote.graph import (
    largest_connected_component_order,
    second_largest_connected_component_order,
    check_graph,
)


def floatsam_threshold_learner(
    graph: AnyGraph,
    starting_treshold: float = 0.0,
    learning_rate: float = 0.01,
    max_drifter_order: Optional[int] = None,
    edge_weight_attr: str = "weight",
):
    """
    Function using an iterative algorithm to try and find the best weight
    threshold to apply to trim the given graph's edges while keeping the
    underlying community structure.

    It works by iteratively increasing the threshold and stopping as soon as
    a significant connected component starts to drift away from the principal
    one.

    This is basically an optimization algorithm applied to a complex nonlinear
    function using a very naive cost heuristic, but it works decently for typical
    cases as it emulates the method used by hand by some researchers when they
    perform this kind of task on Gephi, for instance.

    Args:
        graph (nx.Graph): Graph to sparsify.
        starting_treshold (float, optional): Starting similarity threshold.
            Defaults to `0.0`.
        learning_rate (float, optional): How much to increase the threshold
            at each step of the algorithm. Defaults to `0.05`.
        max_drifter_order (int, optional): Max order of component to detach itself
            from the principal one before stopping the algorithm. If not
            provided it will default to the logarithm of the graph's largest
            connected component's order.
        edge_weight_attr (str, optional): Name of the weight attribute.
            Defaults to "weight".

    Returns:
        float: The found threshold
    """
    check_graph(graph)

    if graph.size() == 0:
        return starting_treshold

    threshold = starting_treshold
    best_threshold = None

    def edge_filter(u, v, attr):
        return attr[edge_weight_attr] >= threshold

    if max_drifter_order is None:
        max_drifter_order = int(math.log(largest_connected_component_order(graph) or 1))

    while True:
        best_threshold = threshold
        threshold += learning_rate

        c = second_largest_connected_component_order(graph, edge_filter)

        if c is not None and c >= max_drifter_order:
            break

    return best_threshold



# =============================================================================
# Pelote Library Endpoint
# =============================================================================
#
from pelote.graph import (
    largest_connected_component,
    crop_to_largest_connected_component,
    remove_edges,
    filter_edges,
)
from pelote.graph_to_tabular import (
    graph_to_nodes_dataframe,
    graph_to_edges_dataframe,
    graph_to_dataframes,
)
from pelote.learn import floatsam_threshold_learner
from pelote.metrics import edge_disparity
from pelote.projection import monopartite_projection
from pelote.read import read_graphology_json
from pelote.sparsification import global_threshold_sparsify
from pelote.tabular_to_graph import table_to_bipartite_graph

__all__ = [
    "largest_connected_component",
    "crop_to_largest_connected_component",
    "remove_edges",
    "filter_edges",
    "graph_to_nodes_dataframe",
    "graph_to_edges_dataframe",
    "graph_to_dataframes",
    "floatsam_threshold_learner",
    "edge_disparity",
    "monopartite_projection",
    "read_graphology_json",
    "global_threshold_sparsify",
    "table_to_bipartite_graph",
]



from pelote.graph_to_tabular import (
    graph_to_nodes_dataframe,
    graph_to_edges_dataframe,
    graph_to_dataframes,
)
from pelote.graph import (
    largest_connected_component,
    crop_to_largest_connected_component,
    remove_edges,
    filter_edges,
)
from pelote.learn import floatsam_threshold_learner
from pelote.metrics import edge_disparity
from pelote.projection import monopartite_projection
from pelote.read import read_graphology_json
from pelote.sparsification import global_threshold_sparsify
from pelote.tabular_to_graph import table_to_bipartite_graph

DOCS = [
    {"title": "Tabular data to graphs", "fns": [table_to_bipartite_graph]},
    {
        "title": "Graphs to tabular data",
        "fns": [
            graph_to_nodes_dataframe,
            graph_to_edges_dataframe,
            graph_to_dataframes,
        ],
    },
    {"title": "Graph projection", "fns": [monopartite_projection]},
    {"title": "Graph sparsification", "fns": [global_threshold_sparsify]},
    {"title": "Miscellaneous graph-related metrics", "fns": [edge_disparity]},
    {
        "title": "Graph utilities",
        "fns": [
            largest_connected_component,
            crop_to_largest_connected_component,
            remove_edges,
            filter_edges,
        ],
    },
    {"title": "Learning", "fns": [floatsam_threshold_learner]},
    {"title": "Reading & Writing", "fns": [read_graphology_json]},
]

__all__ = ["DOCS"]



import re
from io import StringIO
from functools import partial
from docstring_parser import parse as docstring_parser
from docstring_parser import (
    DocstringParam,
    DocstringReturns,
    DocstringRaises,
    DocstringDeprecated,
)

from scripts.docs import DOCS


with open("./README.template.md") as f:
    TEMPLATE = f.read()


def build_toc(data):
    lines = []

    for item in data:
        lines.append(
            "* [%s](#%s)" % (item["title"], item["title"].lower().replace(" ", "-"))
        )

        for fn in item["fns"]:
            name = fn.__name__

            lines.append("  * [%s](#%s)" % (name, name.lower()))

    return "\n".join(lines)


def assembling_description(docstring):
    d = docstring.short_description or ""

    if docstring.long_description:
        if docstring.blank_after_long_description:
            d += "\n" + docstring.long_description
        else:
            d += " " + docstring.long_description

    return d.strip()


EXAMPLE_BLACKLIST = (
    DocstringParam,
    DocstringReturns,
    DocstringRaises,
    DocstringDeprecated,
)


def examples_iter(docstring):
    for meta in docstring.meta:
        if isinstance(meta, EXAMPLE_BLACKLIST):
            continue

        yield meta.description


DEFAULT_RE = re.compile(r".\s*Defaults? to (.+)\.", re.MULTILINE)
CLEAN_RE = re.compile(r"`")


def template_param(param):
    line = "* **%s** " % param.arg_name

    if param.type_name:
        if param.is_optional:
            line += "*%s, optional*" % param.type_name
        else:
            line += "*%s*" % param.type_name

    m = DEFAULT_RE.search(param.description)

    if m is not None:
        line += " `%s`" % CLEAN_RE.sub("", m.group(1))

    line += " - %s" % DEFAULT_RE.sub(".", param.description)

    return line


def template_return(param):
    line = "*%s*" % param.type_name
    line += " - %s" % DEFAULT_RE.sub(".", param.description)
    return line


def build_docs(data):
    f = StringIO()

    p = partial(print, file=f)

    p()

    for item in data:
        p()
        p("---")
        p()
        p("### %s" % item["title"])
        p()

        for fn in item["fns"]:
            name = fn.__name__
            docstring = docstring_parser(fn.__doc__)
            description = assembling_description(docstring)

            p("#### %s" % name)
            p()
            p(description)

            for example in examples_iter(docstring):
                p()
                p("```python")
                p(example)
                p("```")

            p()
            p("*Arguments*")
            p()
            for param in docstring.params:
                p(template_param(param))

            p()

            if docstring.returns:
                p("*Yields*" if docstring.returns.is_generator else "*Returns*")
                p()
                p(template_return(docstring.returns))
                p()

    result = f.getvalue()
    f.close()

    return result


readme = TEMPLATE.format(toc=build_toc(DOCS), docs=build_docs(DOCS)).rstrip()

print(readme)



# =============================================================================
# Pelote Network to Tabular Conversion Functions
# =============================================================================
#
# Functions able to convert networkx graphs to various tabular data formats.
#
from typing import Optional, Tuple, Mapping, Iterable, Union, Dict, Any

from pelote.types import AnyGraph
from pelote.shim import pd, check_pandas
from pelote.graph import check_graph


def graph_to_nodes_dataframe(
    graph: AnyGraph, node_key_col: Optional[str] = "key"
) -> "pd.DataFrame":
    """
    Function converting the given networkx graph into a pandas DataFrame of
    its nodes.

    Args:
        nx.AnyGraph: a networkx graph instance
        node_key_col (str, optional): name of the DataFrame column containing
            the node keys. If None, the node keys will be used as the DataFrame
            index. Defaults to "key".

    Returns:
        pd.DataFrame: A pandas DataFrame

    Example:
        from pelote import to_nodes_dataframe

        df = to_nodes_dataframe(graph)
    """

    check_pandas()
    check_graph(graph)

    if node_key_col is None:

        def raw_data():
            for _, a in graph.nodes(data=True):
                yield a

        return pd.DataFrame(data=raw_data(), index=list(graph.nodes))

    else:

        def data_with_key():
            for n, a in graph.nodes(data=True):
                r = {node_key_col: n}
                r.update(a)
                yield r

        return pd.DataFrame(data=data_with_key())


NodeDataSpec = Union[Mapping[str, str], Iterable[str]]


def graph_to_edges_dataframe(
    graph: AnyGraph,
    *,
    edge_source_col: Optional[str] = "source",
    edge_target_col: Optional[str] = "target",
    source_node_data: Optional[NodeDataSpec] = None,
    target_node_data: Optional[NodeDataSpec] = None
) -> "pd.DataFrame":
    """
    Function converting the given networkx graph into a pandas DataFrame of
    its edges.

    Args:
        nx.AnyGraph: a networkx graph instance
        edge_source_col (str, optional): name of the DataFrame column containing
            the edge source. Defaults to "source".
        edge_target_col (str, optional): name of the DataFrame column containing
            the edge target. Defaults to "target".
        source_node_data (Iterable or Mapping, optional): iterable of attribute names
            or mapping from attribute names to column name to be used to add
            columns to the resulting dataframe based on source node data.
            Defaults to None.
        target_node_data (Iterable or Mapping, optional): iterable of attribute names
            or mapping from attribute names to column name to be used to add
            columns to the resulting dataframe based on target node data.
            Defaults to None.

    Returns:
        pd.DataFrame: A pandas DataFrame
    """

    check_pandas()
    check_graph(graph)

    if source_node_data is not None:
        if not isinstance(source_node_data, (Mapping, Iterable)):
            raise TypeError(
                "source_node_data should be an iterable of keys or a mapping of keys to extract from nodes"
            )

        if not isinstance(source_node_data, Mapping):
            source_node_data = {k: k for k in source_node_data}

    if target_node_data is not None:
        if not isinstance(target_node_data, (Mapping, Iterable)):
            raise TypeError(
                "target_node_data should be an iterable of keys or a mapping of keys to extract from nodes"
            )

        if not isinstance(target_node_data, Mapping):
            target_node_data = {k: k for k in target_node_data}

    def data():
        for source, target, a in graph.edges(data=True):
            r = {edge_source_col: source, edge_target_col: target}
            r.update(a)

            if source_node_data is not None:
                node_data = graph.nodes[source]

                for attr_name, col_name in source_node_data.items():
                    r[col_name] = node_data.get(attr_name)

            if target_node_data is not None:
                node_data = graph.nodes[target]

                for attr_name, col_name in target_node_data.items():
                    r[col_name] = node_data.get(attr_name)

            yield r

    return pd.DataFrame(data=data())


def graph_to_dataframes(
    graph: AnyGraph,
    *,
    node_key_col: Optional[str] = "key",
    edge_source_col: Optional[str] = "source",
    edge_target_col: Optional[str] = "target",
    source_node_data: Optional[NodeDataSpec] = None,
    target_node_data: Optional[NodeDataSpec] = None
) -> Tuple["pd.DataFrame", "pd.DataFrame"]:
    """
    Function converting the given networkx graph into two pandas DataFrames:
    one for its nodes, one for its edges.

    Args:
        nx.AnyGraph: a networkx graph instance
        node_key_col (str, optional): name of the node DataFrame column containing
            the node keys. If None, the node keys will be used as the DataFrame
            index. Defaults to "key".
        edge_source_col (str, optional): name of the edge DataFrame column containing
            the edge source. Defaults to "source".
        edge_target_col (str, optional): name of the edge DataFrame column containing
            the edge target. Defaults to "target".
        source_node_data (Iterable or Mapping, optional): iterable of attribute names
            or mapping from attribute names to column name to be used to add
            columns to the edge dataframe based on source node data.
            Defaults to None.
        target_node_data (Iterable or Mapping, optional): iterable of attribute names
            or mapping from attribute names to column name to be used to add
            columns to the edge dataframe based on target node data.
            Defaults to None.

    Returns:
        (pd.DataFrame, pd.DataFrame)
    """
    nodes = graph_to_nodes_dataframe(graph, node_key_col=node_key_col)
    edges = graph_to_edges_dataframe(
        graph,
        edge_source_col=edge_source_col,
        edge_target_col=edge_target_col,
        source_node_data=source_node_data,
        target_node_data=target_node_data,
    )

    return nodes, edges



# =============================================================================
# Pelote Sparsification Methods
# =============================================================================
#
# Collection of sparsification methods & helpers.
#
from typing import Optional

from pelote.types import AnyGraph
from pelote.graph import remove_edges


def global_threshold_sparsify(
    graph: AnyGraph,
    threshold: float,
    edge_weight_attr: str = "weight",
    reverse: Optional[bool] = False,
) -> None:
    """
    Function sparsifying a networkx graph by removing all its edges
    having a weight less than a given threshold.

    Note that this function mutates the given graph.

    Args:
        graph (nx.AnyGraph): target graph.
        threshold (float): weight threshold.
        reverse (bool, optional): whether to reverse the threshold condition.
            That is to say an edge would be removed if its weight is greater
            than the threshold.
    """

    if reverse:

        def edge_predicate(u, v, a):
            return a[edge_weight_attr] <= threshold

    else:

        def edge_predicate(u, v, a):
            return a[edge_weight_attr] >= threshold

    remove_edges(graph, edge_predicate)



from typing import Any
from contextlib import contextmanager

try:
    import pandas

    pd = pandas
    original_pandas = pandas
except ImportError:
    pd = None
    original_pandas = None


from pelote.exceptions import MissingPandasException


def obliterate_pandas():
    """
    Function forcing the library to consider that `pandas` was not installed,
    which can be useful when unit testing.
    """

    global pd
    pd = None


def resurrect_pandas():
    """
    Function forcing the library to acknowledge that `pandas` is back to life,
    which can be useful when unit testing.
    """

    global pd, original_pandas

    if original_pandas is not None:
        pd = original_pandas


@contextmanager
def missing_pandas():
    obliterate_pandas()

    try:
        yield
    finally:
        resurrect_pandas()


def is_pandas_available() -> bool:
    """
    Function returning whether `pandas` is installed.
    """
    return pd is not None


def check_pandas() -> None:
    """
    Function raising if `pandas` is not installed.

    Note that I could use a decorator but I don't because those cannot be
    usefully typed before python v3.10.
    """
    if pd is None:
        raise MissingPandasException(
            "pandas must be installed for this function to work"
        )


def is_dataframe(value: Any) -> bool:
    global pd

    if pd is None:
        return False

    return isinstance(value, pd.DataFrame)



# =============================================================================
# Pelote Reading Functions
# =============================================================================
#
# Functions used to read from various data formats.
#
import json
import networkx as nx
from pathlib import Path
from io import IOBase
from typing import Union

from pelote.types import GraphologySerializedGraph, AnyGraph, FileHandle


def parse_graphology_json(data: GraphologySerializedGraph) -> AnyGraph:
    if "options" not in data:
        raise TypeError(
            "cannot parse a graphology json that does not record the graph options"
        )

    options = data["options"]
    graph_type = options["type"]

    if graph_type == "mixed":
        raise TypeError("cannot parse a mixed graph yet")

    is_multi = options["multi"]

    graph: AnyGraph

    if graph_type == "directed":
        if is_multi:
            graph = nx.MultiDiGraph()
        else:
            graph = nx.DiGraph()
    else:
        if is_multi:
            graph = nx.MultiGraph()
        else:
            graph = nx.Graph()

    nodes = data.get("nodes")

    if nodes is not None:
        for serialized_node in nodes:
            attr = serialized_node.get("attributes", {})
            graph.add_node(serialized_node["key"], **attr)

    edges = data.get("edges")

    if edges is not None:
        for serialized_edge in edges:
            key = serialized_edge.get("key")
            attr = serialized_edge.get("attributes", {})
            source = serialized_edge["source"]
            target = serialized_edge["target"]

            if key is not None and is_multi:
                graph.add_edge(source, target, key=key, **attr)
            else:
                graph.add_edge(source, target, **attr)

    return graph


def read_graphology_json(
    target: Union[FileHandle, GraphologySerializedGraph]
) -> AnyGraph:
    """
    Function reading and parsing the given json file as a networkx graph.

    Args:
        target (str or Path or file or dict): target to read and parse. Can
            be a string path, a Path instance, a file buffer or already
            parsed JSON data as a dict.

    Returns:
        nx.AnyGraph: a networkx graph instance.
    """
    data: GraphologySerializedGraph

    if isinstance(target, (str, Path)):
        with open(target) as f:
            data = json.load(f)

    elif isinstance(target, IOBase):
        data = json.load(target)

    elif isinstance(target, dict):
        data = target

    else:
        raise TypeError("expected a path or a file")

    return parse_graphology_json(data)



# =============================================================================
# Pelote Utilities
# =============================================================================
#
# Miscellaneous utility functions used throughout the library.
#
from typing import Iterable, Any, Dict, TypeVar, Generic, List, Set, Optional

from pelote.types import AnyGraph

K = TypeVar("K")
V = TypeVar("V")


def has_mixed_types(iterable: Iterable[Any]) -> bool:
    iterator = iter(iterable)

    main_type = None

    for item in iterator:
        main_type = type(item)
        break

    for item in iterator:
        if type(item) is not main_type:
            return True

    return False


class IncrementalIdRegister(Generic[K]):
    def __init__(self):
        self.__counter = 0
        self.__index: Dict[K, int] = {}

    def __getitem__(self, item: K) -> int:
        item_id = self.__index.get(item)

        if item_id is None:
            item_id = self.__counter
            self.__counter += 1
            self.__index[item] = item_id

        return item_id


class DFSStack(Generic[V]):
    def __init__(self, graph: AnyGraph):
        self.__order = graph.order()
        self.__stack: List[V] = []
        self.__seen: Set[V] = set()

    def __len__(self) -> int:
        return len(self.__stack)

    def __contains__(self, node: V) -> bool:
        return node in self.__seen

    def has_seen_everything(self) -> bool:
        return len(self.__seen) == self.__order

    def append(self, node: V) -> bool:
        size_before = len(self.__seen)

        self.__seen.add(node)

        if size_before == len(self.__seen):
            return False

        self.__stack.append(node)

        return True

    def pop(self) -> Optional[V]:
        if len(self.__stack) == 0:
            return None

        return self.__stack.pop()


def dict_without(d: Dict[K, V], k: str) -> Dict[K, V]:
    o = {}

    for n, v in d.items():
        if n == k:
            continue

        o[n] = v

    return o



# =============================================================================
# Pelote Metrics Functions
# =============================================================================
#
from typing import Dict, Any, Optional

from pelote.types import AnyGraph


def edge_disparity(
    graph: AnyGraph, edge_weight_attr: str = "weight"
) -> Dict[Any, float]:
    """
    Function computing the disparity score of each edge in the given graph. This
    score is typically used to extract the multiscale backbone of a weighted
    graph.

    Args:
        graph(nx.AnyGraph): target graph.
        edge_weight_attr (str, optional): name of the edge attribute containing
            its weight.
            Defaults to "weight".

    Returns:
        dict: Dictionnary with edges - (source, target) tuples - as keys and the disparity scores as values

    """
    # todo: raise if multigraph, raise if directed or at least change code to optimize

    disparities = {}
    previous: Optional[Any] = None
    weighted_degrees = graph.degree(weight=edge_weight_attr)

    for source, target, weight in graph.edges.data(data=edge_weight_attr):

        if previous is None or previous != source:
            previous = source
            previous_degree = graph.degree(source)
            previous_weighted_degree = weighted_degrees[source]

        target_degree = graph.degree(target)
        target_weighted_degree = weighted_degrees[target]

        normalized_weight_source = weight / previous_weighted_degree
        normalized_weight_target = weight / target_weighted_degree

        source_score = (1 - normalized_weight_source) ** (previous_degree - 1)
        target_score = (1 - normalized_weight_target) ** (target_degree - 1)

        disparities[(source, target)] = min(source_score, target_score)

    return disparities



import networkx as nx
from typing import Any, Union, Dict, List, Iterable
from typing_extensions import TypedDict, Literal, NotRequired, Protocol
from pathlib import Path
from io import IOBase

from pelote.shim import pd

# Misc utils
FileHandle = Union[str, Path, IOBase]
GenericKey = Union[str, int]


class Indexable(Protocol):
    def __getitem__(self, key: Any) -> Any:
        ...


Tabular = Union["pd.DataFrame", Iterable[Indexable]]

# Networkx-related
AnyGraph = Union[
    nx.Graph,
    nx.DiGraph,
    nx.MultiGraph,
    nx.MultiDiGraph,
]
Attributes = Dict[Any, Any]

# Graphology serialization
SerializedAttributes = Dict[str, Any]
GraphologyType = Literal["directed", "undirected", "mixed"]


class GraphologySerializedNode(TypedDict):
    key: str
    attributes: NotRequired[SerializedAttributes]


class GraphologySerializedEdge(TypedDict):
    key: NotRequired[str]
    source: str
    target: str
    attributes: NotRequired[SerializedAttributes]
    undirected: NotRequired[bool]


class GraphologySerializedOptions(TypedDict):
    allowSelfLoops: bool
    type: GraphologyType
    multi: bool


class GraphologySerializedGraph(TypedDict):
    options: NotRequired[GraphologySerializedOptions]
    attributes: NotRequired[SerializedAttributes]
    nodes: NotRequired[List[GraphologySerializedNode]]
    edges: NotRequired[List[GraphologySerializedEdge]]



# =============================================================================
# Pelote Graph Helpers
# =============================================================================
#
# Miscellaneous helper functions to deal with networkx graphs.
#
import networkx as nx
from heapq import nlargest
from typing import Set, Any, Optional, List, Callable, Dict, Generator
from typing_extensions import TypeGuard

from pelote.types import AnyGraph, Attributes
from pelote.utils import DFSStack

GRAPH_TYPES = (nx.Graph, nx.DiGraph, nx.MultiGraph, nx.MultiDiGraph)


def is_graph(value: Any) -> TypeGuard[AnyGraph]:
    """
    Function returning whether the given value is a networkx graph.

    Args:
        value (any): value to test.

    Returns
        bool
    """
    return isinstance(value, GRAPH_TYPES)


def check_graph(value: Any) -> None:
    """
    Function raising if the given value is not a networkx graph.

    Args:
        value (any): value to test.
    """
    if not is_graph(value):
        raise TypeError("expected a networkx graph but got %s" % type(value).__name__)


def largest_connected_component(graph: AnyGraph) -> Optional[Set[Any]]:
    """
    Function returning the largest connected component of given networkx graph
    as a set of nodes.

    Note that this function will consider any given graph as undirected and
    will therefore work with weakly connected components in the directed case.

    Args:
        graph (nx.AnyGraph): target graph.

    Returns:
        set: set of nodes representing the largest connected component.
    """
    check_graph(graph)

    largest = None
    remaining_nodes = graph.order()

    components = (
        nx.connected_components
        if not graph.is_directed()
        else nx.weakly_connected_components
    )

    for component in components(graph):
        if largest is None or len(component) > len(largest):
            largest = component

        # Early exit
        remaining_nodes -= len(largest)

        if len(largest) > remaining_nodes:
            break

    return largest


def crop_to_largest_connected_component(graph: AnyGraph) -> None:
    """
    Function mutating the given networkx graph in order to keep only the
    largest connected component.

    Note that this function will consider any given graph as undirected and
    will therefore work with weakly connected components in the directed case.

    Args:
        graph (nx.AnyGraph): target graph.
    """
    check_graph(graph)

    component = largest_connected_component(graph)

    if component is None:
        return

    nodes_to_drop: List[Any] = []

    for node in graph:
        if node not in component:
            nodes_to_drop.append(node)

    for node in nodes_to_drop:
        graph.remove_node(node)


def have_same_nodes(A: AnyGraph, B: AnyGraph, check_attributes: bool = False) -> bool:
    if A.order() != B.order():
        return False

    for node, attr in A.nodes.data():
        if node not in B:
            return False

        if check_attributes and attr != B.nodes[node]:
            return False

    return True


def have_same_edges(A: AnyGraph, B: AnyGraph, check_attributes: bool = False) -> bool:
    if A.size() != B.size():
        return False

    for u, v, attr in A.edges.data():
        if not B.has_edge(u, v):
            return False

        if check_attributes and attr != B[u][v]:
            return False

    return True


def are_same_graphs(A: AnyGraph, B: AnyGraph, check_attributes: bool = False) -> bool:
    return have_same_nodes(A, B, check_attributes=check_attributes) and have_same_edges(
        A, B, check_attributes=check_attributes
    )


EdgePredictate = Callable[[Any, Any, Dict[Any, Any]], bool]


def remove_edges(graph: AnyGraph, predicate: EdgePredictate) -> None:
    """
    Function removing all edges that do not pass a predicate function from a
    given networkx graph.

    Note that this function mutates the given graph.

    Args:
        graph (nx.AnyGraph): a networkx graph.
        predicate (callable): a function taking each edge source, target and
            attributes and returning True if you want to keep the edge or False
            if you want to remove it.
    """
    check_graph(graph)

    if not callable(predicate):
        raise TypeError("expecting a callable predicate (i.e. a function etc.)")

    edges_to_drop = []

    for u, v, a in graph.edges.data():
        if not predicate(u, v, a):
            edges_to_drop.append((u, v))

    for u, v in edges_to_drop:
        graph.remove_edge(u, v)


def filter_edges(graph: AnyGraph, predicate: EdgePredictate) -> AnyGraph:
    """
    Function returning a copy of the given networkx graph but without the edges
    filtered out by the given predicate function

    Args:
        graph (nx.AnyGraph): a networkx graph.
        predicate (callable): a function taking each edge source, target and
            attributes and returning True if you want to keep the edge or False
            if you want to remove it.

    Returns:
        nx.AnyGraph: the filtered graph.
    """
    check_graph(graph)

    if not callable(predicate):
        raise TypeError("expecting a callable predicate (i.e. a function etc.)")

    copy = nx.create_empty_copy(graph)

    for u, v, a in graph.edges.data():
        if predicate(u, v, a):
            copy.add_edge(u, v, **a)

    return copy


EdgeFilterFunction = Callable[[Any, Any, Attributes], bool]


def connected_component_orders(
    graph: AnyGraph,
    edge_filter: Optional[EdgeFilterFunction] = None,
) -> Generator[int, None, None]:
    """
    Function yielding the given graph's connected component orders. It is
    faster than calling `len` on sets yielded by nx.connected_components and
    can use an edge filter function.

    Args:
        graph (nx.AnyGraph): a networkx graph.
        edge_filter (callable, optional): a function taking n1, n2 & the
            attributes and returning whether we should follow this edge or not.
            Defaults to None.

    Yields:
        int: the size of a connected component.
    """
    check_graph(graph)

    if edge_filter is not None and not callable(edge_filter):
        raise TypeError("edge_filter should be callable")

    # Wrapping generator to make sure type checking raises on call
    def generator():
        stack = DFSStack[Any](graph)

        for node in graph:
            if stack.has_seen_everything():
                break

            if node in stack:
                continue

            stack.append(node)
            size = 0

            while len(stack) != 0:
                n1 = stack.pop()
                size += 1

                for _, n2, edge_attr in graph.edges(n1, data=True):
                    if edge_filter is not None and not edge_filter(n1, n2, edge_attr):
                        continue

                    stack.append(n2)

            yield size

    return generator()


def largest_connected_component_order(graph: AnyGraph) -> Optional[int]:
    """
    Function returning the order of the largest connected component of the given
    graph.

    Args:
        graph (nx.AnyGraph): target graph.

    Returns:
        int or None: order of the component or None if the graph is null.
    """
    if graph.order() == 0:
        return None

    return max(connected_component_orders(graph))


def second_largest_connected_component_order(
    graph: AnyGraph, edge_filter: Optional[EdgeFilterFunction] = None
) -> Optional[int]:
    """
    Function returning the order of the second largest connected component
    of the given graph.

    Args:
        graph (nx.AnyGraph): target graph.
        edge_filter (callable, optional): a function taking n1, n2 & the
            attributes and returning whether we should follow this edge or not.
            Defaults to None.

    Returns:
        int or None: order of the component or None if the graph is null or has
            only a single component.
    """
    top2 = nlargest(2, connected_component_orders(graph, edge_filter))

    if len(top2) < 2:
        return None

    return top2[1]
